#region Copyright
// Copyright Syncfusion Inc. 2001 - 2009. All rights reserved.
// Use of this code is subject to the terms of our license.
// A copy of the current license can be obtained at any time by e-mailing
// licensing@syncfusion.com. Any infringement will be prosecuted under
// applicable laws. 
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Drawing;
using System.Reflection;
using Syncfusion.Web.UI.WebControls.Grid.Grouping;

namespace TreeGridListHelper_cs
{
    #region the TreeGridListHelper class

    /// <summary>
    /// Summary description for TreeGridListHelper.
    /// </summary>
    public class TreeGridList
    {

        private SelfReferenceDataSourceAccessor accessor;
        Color baseColor = Color.White;//Color.FromArgb(240, 240, 240);
        private bool colorCodeSiblings = false;
        int colorOffset = 20;
        private ITreeGridDataSource dataSource;
        private bool editingEnabled = false;

        private GridGroupingControl grid = null;
        private string idProperty;
        private int maxIndent = int.MaxValue;
        private string nodeProperty;


        private NodeArray rowIndexer = null;

        private int rowOffSet = 2;
        private string selfReferenceProperty;

        private bool showCellBorderInTreeCell = false;
        private string sortColumn = "";

        private ListSortDirection sortDirection = ListSortDirection.Ascending;

        private const int TREECOLUMN = 1;
        private bool treeLines = true;
        private bool treeNodeEditable = true;



        Dictionary<int, List<bool>> visibleDescendantExpandStates = null;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="dataSource">The DataTable with a column that references to primary key values for the same DataTable. This means there is a field that refers to another DataRow in the same DataTable.</param>
        /// <param name="selfReferenceColumn">The is the name of the column that refers to the primary key value that identifies another DataRow in the DataTabale.</param>
        /// <param name="idColumn">The primary key column for the DataTable.</param>
        /// <param name="nodeColumn">The name of the column that should appear in the +/- expansion cell.</param>
        public TreeGridList(DataTable dataSource, string selfReferenceColumn, string idColumn, string nodeColumn)
        {
            this.selfReferenceProperty = selfReferenceColumn;
            this.idProperty = idColumn;
            this.nodeProperty = nodeColumn;

            this.rowIndexer = new NodeArray(dataSource.DefaultView);

            this.accessor = new SelfReferenceDataSourceAccessor(dataSource, selfReferenceColumn, idColumn);

            this.dataSource = this.accessor;

            rowIndexer.Clear();
            PopulateIndexer(0, accessor.GetRootDataView(), 0);
        }


        public bool ColorCodeSiblings
        {
            get { return colorCodeSiblings; }
            set { colorCodeSiblings = value; }
        }

        /// <summary>
        /// Gets or sets whether non treenode cells are allowed to go into edit mode.
        /// </summary>
        public bool EditingEnabled
        {
            get
            {
                return editingEnabled;
            }
            set
            {
                editingEnabled = value;
            }
        }

        /// <summary>
        /// Gets whether there is a GridDataBoundGrid wired to this TreeHelper.
        /// </summary>
        public bool IsWired
        {
            get
            {
                return this.grid != null;
            }
        }




        /// <summary>
        /// Gets or sets maximum tree levels displayable in the grid. The default is int.MaxValue.
        /// </summary>
        public int MaxIndentLevels
        {
            get { return this.maxIndent; }
            set { this.maxIndent = value; }
        }

        public bool ShowCellBorderInTreeCell
        {
            get { return showCellBorderInTreeCell; }
            set { showCellBorderInTreeCell = value; }
        }

        /// <summary>
        /// Gets or sets whether tree lines are displayed
        /// </summary>
        public bool ShowTreeLines
        {
            get
            {
                return treeLines;
            }
            set
            {
                if (treeLines != value)
                {
                    treeLines = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets whether treenode cell is editable.
        /// </summary>
        public bool TreeNodeCellEditable
        {
            get { return treeNodeEditable; }
            set { treeNodeEditable = true; }
        }

        protected Dictionary<int, List<bool>> VisibleDescendantExpandStates
        {
            get
            {
                if (visibleDescendantExpandStates == null)
                    visibleDescendantExpandStates = new Dictionary<int, List<bool>>();
                return visibleDescendantExpandStates;
            }
            set { visibleDescendantExpandStates = value; }
        }

        /// <summary>
        /// Collapses all the nodes in the grid.
        /// </summary>
        public void CollapseAll()
        {
            try
            {
                int rowIndex = rowOffSet;
                Node n = GetNode(rowIndex);
                IList list = n.SiblingList;
                CollapseNode(rowIndex);
                if (list != null)
                {
                    for (int i = 1; i < list.Count; ++i)
                    {
                        CollapseNode(rowIndex + i);
                    }
                }
            }
            catch { }
            finally
            {
            }
            VisibleDescendantExpandStates.Clear();
        }

        private void CollapseNode(int rowIndex, bool saveExpandStates)
        {
            Node n = GetNode(rowIndex);
            if (n.IsExpanded && n.HasChildren)
            {
                if (saveExpandStates)
                {
                    int id = GetPrimaryId(rowIndex);
                    if (VisibleDescendantExpandStates.ContainsKey(id))
                    {
                        VisibleDescendantExpandStates[id] = QueryVisibleDescendantsExpandStates(rowIndex);
                    }
                    else
                    {
                        VisibleDescendantExpandStates.Add(id, QueryVisibleDescendantsExpandStates(rowIndex));
                    }
                }
                int i = rowIndex + 1;
                int pos = GetPosition(i);
                int rowCount = this.rowIndexer.Count + 1;
                if (i <= rowCount)
                {
                    Node next = GetNode(i);
                    while (next.Indent > n.Indent && i <= rowCount)
                    {
                        this.rowIndexer.RemoveAt(pos);
                        rowCount--;
                        if (i <= rowCount)
                            next = GetNode(i);
                    }
                }
                n.IsExpanded = false;
            }

        }



        /// <summary>
        /// Collapses the node at the grid rowIndex.
        /// </summary>
        /// <param name="rowIndex">The rowIndex.</param>
        public void CollapseNode(int rowIndex)
        {
            CollapseNode(rowIndex, false);
        }

        /// <summary>
        /// Expands all nodes in the wired grid.
        /// </summary>
        public void ExpandAll()
        {
            this.CollapseAll(); //speeds things up...
            int rowIndex = rowOffSet;
            while (rowIndex <= this.rowIndexer.Count + 1)
            {
                ExpandNode(rowIndex, false, false);
                rowIndex += 1;
            }
        }

        private void ExpandDescendants(int rowIndex)
        {
            //int row = rowIndex + 1;
            //List<bool> expandStates = GetVisibleDescendantExpandStates(rowIndex);
            //foreach (bool b in expandStates)
            //{
            //    if (b && row < grid.TableModel.RowCount)
            //    {
            //        ExpandNode(row, false, false, false);
            //    }
            //    row++;
            //}
        }


        private void ExpandNode(int rowIndex, bool refresh, bool resizeNodeCell)
        {
            ExpandNode(rowIndex, refresh, resizeNodeCell, false);
        }

        private void ExpandNode(int rowIndex, bool refresh, bool resizeNodeCell, bool expandDescendents)
        {
            Node n = GetNode(rowIndex);

            if (n.Indent > this.MaxIndentLevels)
                return;



            if (n.HasChildren)
            {
                //Console.WriteLine(rowIndex);
                try
                {
                    int pos = GetPosition(rowIndex);
                    IList list = (n.ChildList == null) ? this.dataSource.GetChildList(n.Position, n.SiblingList, n.ParentNode) : n.ChildList;

                    PopulateIndexer(pos + 1, list, n.Indent + 1, n);
                    n.IsExpanded = true;
                    if (expandDescendents)
                    {
                        ExpandDescendants(rowIndex);
                        int id = GetPrimaryId(rowIndex);
                        if (VisibleDescendantExpandStates.ContainsKey(id))
                        {
                            VisibleDescendantExpandStates.Remove(id);
                        }
                    }
                }
                catch { }
                finally
                {
                }
            }
        }


        /// <summary>
        /// Expands the tree node at grid rowIndex.
        /// </summary>
        /// <param name="rowIndex">The rowIndex of the node to be expanded.</param>
        /// <param name="refresh">Indicates whether grid.Refresh is called after the expansion.</param>
        public void ExpandNode(int rowIndex, bool refresh)
        {
            ExpandNode(rowIndex, refresh, true);
        }

        /// <summary>
        /// Expands a single node at grid rowIndex.
        /// </summary>
        /// <param name="rowIndex">The row index of the node to be expanded.</param>
        public void ExpandNode(int rowIndex)
        {
            ExpandNode(rowIndex, true);
        }



        private List<int> GetExpandedNodes()
        {
            List<int> expandedNodes = new List<int>();
            //if (rowIndexer.Count > 0)
            //{
            //    for (int row = rowOffSet; row <= grid.TableModel.RowCount; ++row)
            //    {
            //        Node n = rowIndexer[row - rowOffSet];
            //        if (n.IsExpanded)
            //        {
            //            int id = GetPrimaryId(row);
            //            int loc = expandedNodes.BinarySearch(id);
            //            if (loc < 0)
            //                expandedNodes.Insert(-(loc + 1), id);
            //        }
            //    }
            //}
            return expandedNodes;
        }

        private Color GetIndentLevelColor(int indent)
        {
            return Color.FromArgb(Math.Max(0, baseColor.R - indent * colorOffset), Math.Max(0, baseColor.G - indent * colorOffset), Math.Max((ushort)0, baseColor.B));//- indent * 10); 
        }

        /// <summary>
        /// Returns the Node object for the given grid rowIndex.
        /// </summary>
        /// <param name="rowIndex">The rowIndex.</param>
        /// <returns>The Node for the given row.</returns>
        public Node GetNode(int rowIndex)
        {
            int pos = GetPosition(rowIndex);
            return this.rowIndexer[pos];
        }

        private int GetPosition(int rowIndex)
        {
            return rowIndex - rowOffSet;
        }


        private int GetPrimaryId(int rowIndex)
        {
            if (rowIndex >= rowOffSet && rowIndex <= this.rowIndexer.Count + 1
                && this.rowIndexer.Count > 0)
            {
                int pos = GetPosition(rowIndex);// this.grid.TableControl.Binder.RowIndexToPosition(rowIndex);
                Node n = this.rowIndexer[pos];
                DataRowView drv = n.SiblingList[n.Position] as DataRowView;
                if (drv != null)
                {
                    return (int)drv[idProperty];
                }
            }
            return -1;
        }

        private List<bool> GetVisibleDescendantExpandStates(int rowIndex)
        {
            int i = GetPrimaryId(rowIndex);
            if (!VisibleDescendantExpandStates.ContainsKey(i))
            {
                VisibleDescendantExpandStates.Add(i, new List<bool>());
            }
            return VisibleDescendantExpandStates[i];
        }

        private void grid_CurrentCellStartEditing(object sender, CancelEventArgs e)
        {
            e.Cancel = !this.editingEnabled;
        }

        void grid_QueryCellStyleInfo(object sender, GridTableCellStyleInfoEventArgs e)
        {

            if (e.Style.TableCellIdentity.TableCellType == GridTableCellType.ColumnHeaderCell
                && e.Style.TableCellIdentity.Column != null && e.Style.TableCellIdentity.Column.Name == sortColumn)
            {
                e.Style.Tag = sortDirection;
            }
        }

        private void PopulateIndexer(int pos, IList data, int indent, Node parent)
        {
            for (int i = 0; i < data.Count; ++i)
            {
                DataRowView drv = data[i] as DataRowView;
                Node node1 = new Node(drv);
                node1.Position = i;
                node1.SiblingList = data;
                node1.Indent = indent;
                node1.ParentNode = parent;
                node1.HasChildren = this.dataSource.NodeHasChildren(i, data);
                this.rowIndexer.Insert(pos, node1);
                pos += 1;
            }
        }

        private void PopulateIndexer(int pos, IList data, int indent)
        {
            PopulateIndexer(pos, data, indent, null);
        }

        private List<bool> QueryVisibleDescendantsExpandStates(int rowIndex)
        {
            List<bool> expandStates = new List<bool>();
            int indent = GetNode(rowIndex).Indent;
            int row = rowIndex + 1;
            bool continueProcessing = true;
            while (continueProcessing && row < this.rowIndexer.Count + 1) //no need to check last row as it is nochild or collapsed
            {
                Node n = GetNode(row);
                if (n.Indent > indent)
                {
                    expandStates.Add(n.HasChildren && n.IsExpanded);
                }
                else
                    continueProcessing = false;
                row++;
            }

            return expandStates;
        }

        private void ResetExpandedNodes(List<int> expandedNodes)
        {
            //grid.BeginUpdate();
            int row = 1;
            while (row <= rowIndexer.Count)
            {
                int id = GetPrimaryId(row);
                int loc = expandedNodes.BinarySearch(id);
                if (loc >= 0)
                    ExpandNode(row);
                row++;
            }
            //grid.EndUpdate(true);
        }


        public void Sort(string sortColumn)
        {
            if (sortColumn.Length == 0)
            {
                this.sortColumn = sortColumn;
                accessor.sortString = sortColumn;
            }
            else if (this.sortColumn == sortColumn)
            {
                accessor.sortString = sortColumn + ((sortDirection == ListSortDirection.Ascending) ? " Desc" : "");

                sortDirection = (sortDirection == ListSortDirection.Ascending) ? ListSortDirection.Descending : ListSortDirection.Ascending;
            }
            else
            {
                sortDirection = ListSortDirection.Ascending;
                accessor.sortString = sortColumn;
                this.sortColumn = sortColumn;

            }


            //      CollapseAll();
            List<int> saveExpands = GetExpandedNodes();
            rowIndexer.Clear();
            PopulateIndexer(0, accessor.GetRootDataView(), 0);
            ResetExpandedNodes(saveExpands);
            // ExpandAll();
        }

        /// <summary>
        /// Disconnects this TreeHelper from its GridDataBoundGrid.
        /// </summary>
        public void UnwireGrid()
        {
            if (this.grid != null)
            {
                this.grid.QueryCellStyleInfo -= new GridTableCellStyleInfoEventHandler(grid_QueryCellStyleInfo);

                this.grid = null;
            }
        }

        /// <summary>
        /// Does some low level initializations. Generally, you should use the WireSelfReferenceGrid method which, in turn, will 
        /// call WireGrid.
        /// </summary>
        /// <param name="grid">The GridDataBoundGrid</param>
        public void WireGrid(GridGroupingControl grid)
        {
            this.grid = grid;

            WireTreeLineEvents();
        }

        /// <summary>
        /// Adds tree node support to a GridGroupingControl whose datasource is a DataTable with
        /// a self-reference.
        /// </summary>
        /// <param name="grid">The GridGroupingControl.</param>
        public void WireSelfReferenceGrid(GridGroupingControl grid)
        {


            this.grid = grid;
            this.grid.TopLevelGroupOptions.ShowAddNewRecordBeforeDetails = false;
            this.grid.TopLevelGroupOptions.ShowCaption = false;

            this.grid.DataSource = this.rowIndexer;// accessor.GetRootDataView();
            WireGrid(grid);

            //make sure nodeColumn is the first one...
            GridColumnDescriptor gbc = grid.TableDescriptor.Columns.FindByMappingName(this.nodeProperty);
            if (gbc == null)
            {
                throw new Exception(string.Format("{0} is missing from the grid.", this.nodeProperty));
            }
            int loc = grid.TableDescriptor.Columns.IndexOf(gbc);
            if (loc != 0)
            {
                grid.TableDescriptor.Columns.Move(loc, 0);

            }
            foreach (GridColumnDescriptor cd in grid.TableDescriptor.Columns)
            {
                cd.AllowSort = false;
            }
        }

        private void WireTreeLineEvents()
        {
            this.grid.QueryCellStyleInfo += new GridTableCellStyleInfoEventHandler(grid_QueryCellStyleInfo);
        }

        #region private fields
        //Color.FromArgb(240, 240, 240);




        #endregion

        #region Properties

        #endregion



        #region non-public methods methods



        #endregion


    }

    #endregion

    #region The NodeArray and Node classes
    public class NodeArray : BindingList<Node>, ITypedList, IListSource
    {
        private DataView dv;

        public NodeArray(DataView dv)
        {
            this.dv = dv;
        }

        public bool ContainsListCollection
        {
            get { return false; }
        }

        public new Node this[int i]
        {
            get { return (Node)base[i]; }
            set { base[i] = value; }
        }

        public PropertyDescriptorCollection GetItemProperties(PropertyDescriptor[] listAccessors)
        {
            ITypedList list = (ITypedList)dv;


            PropertyDescriptorCollection pdc = ((ITypedList)dv).GetItemProperties(listAccessors);
            CustomPropertyDescriptor[] pds = new CustomPropertyDescriptor[pdc.Count];
            int i = 0;
            foreach (PropertyDescriptor pd in pdc)
            {
                pds[i++] = new CustomPropertyDescriptor(pd);
            }
            return new PropertyDescriptorCollection(pds);
        }

        public IList GetList()
        {
            return this;
        }

        public string GetListName(PropertyDescriptor[] listAccessors)
        {
            return "NodeDrv";
        }

        #region ITypedList Members


        #endregion

        #region IListSource Members


        #endregion
    }

    public class Node
    {
        IList cList = null;

        DataRowView drv;
        bool expanded = false;
        bool hasChildren1 = true;
        int indent1;
        Node pNode = null;
        int pos1;
        IList sList = null;

        public Node(DataRowView drv)
        {
            this.drv = drv;
        }

        public IList ChildList
        {
            get { return cList; }
            set { cList = value; }
        }

        public DataRowView Drv
        {
            get { return drv; }
            set { drv = value; }
        }

        public bool HasChildren
        {
            get { return hasChildren1; }
            set { hasChildren1 = value; }
        }

        public int Indent
        {
            get { return indent1; }
            set { indent1 = value; }
        }

        public bool IsExpanded
        {
            get { return expanded; }
            set { expanded = value; }
        }

        public bool IsInitialized
        {
            get { return !hasChildren1 || cList != null; }
        }


        public Node ParentNode
        {
            get { return pNode; }
            set { pNode = value; }
        }

        public int Position
        {
            get { return pos1; }
            set { pos1 = value; }
        }

        public IList SiblingList
        {
            get { return sList; }
            set { sList = value; }
        }
    }

    #region custom property descriptor  - Used to expose properties in the DataTable as properties in a NodeList member
    /// <summary>
    /// Used to expose properties in the DataTable as properties in a NodeList member.
    /// </summary>
    public class CustomPropertyDescriptor : PropertyDescriptor
    {
        private PropertyDescriptor pd;

        public CustomPropertyDescriptor(PropertyDescriptor pd)
            : base(pd.Name, null)
        {
            this.pd = pd;
        }

        public override Type ComponentType
        {
            get { return pd.ComponentType; }
        }

        public override bool IsReadOnly
        {
            get { return pd.IsReadOnly; }
        }

        public override Type PropertyType
        {
            get { return pd.PropertyType; }
        }

        public override bool CanResetValue(object component)
        {
            return pd.CanResetValue(((Node)component).Drv);
        }

        public override object GetValue(object component)
        {
            if (((Node)component).Drv == null)
                return null;
            return pd.GetValue(((Node)component).Drv);
        }

        public override void ResetValue(object component)
        {
            pd.ResetValue(((Node)component).Drv);
        }

        public override void SetValue(object component, object value)
        {
            pd.SetValue(((Node)component).Drv, value);
        }

        public override bool ShouldSerializeValue(object component)
        {
            return pd.ShouldSerializeValue(((Node)component).Drv);
        }
    }
    #endregion
    #endregion

    #region mainly internal use - support classes for tree structure and datasources

    public interface ITreeGridDataSource
    {
        bool NodeHasChildren(int pos, IList parent);
        bool NodeHasChildren(int pos, IList parent, Node parentNode);
        IList GetChildList(int pos, IList parent);
        IList GetChildList(int pos, IList parent, Node parentNode);

    }

    public class SelfReferenceDataSourceAccessor : ITreeGridDataSource, IDisposable
    {
        private DataTable dt;
        private DataView dv = null;
        private IList lastParent = null;
        private int lastPos;
        private string matchProperty;
        private bool reset1 = false;
        internal string sortString = "";
        private string valueProperty;

        public SelfReferenceDataSourceAccessor(DataTable dt, string matchProperty, string valueProperty)
        {
            this.dt = dt;
            this.matchProperty = matchProperty;
            this.valueProperty = valueProperty;

        }

        public void Dispose()
        {
            if (dv != null)
            {
                dv.Dispose();
                dv = null;
            }

        }

        public DataView GetChildDataView(int pos, IList parent)
        {
            this.lastParent = parent;
            this.lastPos = pos;
            string s = ((DataView)parent)[pos][this.valueProperty].ToString();

            string filter = string.Format("[{0}] = {1}", this.matchProperty, s);


            dv = new DataView(this.dt, filter, sortString, DataViewRowState.CurrentRows);
            //dv = new DataView(this.dt, string.Format("[{0}] = {1} OR [{0}] = {2}", this.matchProperty, s, s1), "", DataViewRowState.CurrentRows);

            return dv;
        }

        public IList GetChildList(int pos, IList parent)
        {
            if (UseSameDataView(pos, parent))
                return this.dv;

            return GetChildDataView(pos, parent);
        }

        public IList GetChildList(int pos, IList parent, Node parentNode)
        {
            if (UseSameDataView(pos, parent))
                return this.dv;

            return GetChildDataView(pos, parent);
        }



        public DataView GetRootDataView()
        {
            string filter = string.Format("[{0}] = -1", matchProperty);
            this.dv = new DataView(this.dt, filter, sortString, DataViewRowState.CurrentRows);
            return this.dv;
        }

        public bool NodeHasChildren(int pos, IList parent, Node parentNode)
        {
            if (UseSameDataView(pos, parent))
                return this.dv.Count > 0;
            this.dv = GetChildDataView(pos, parent);
            return this.dv != null && this.dv.Count > 0;
        }

        public bool NodeHasChildren(int pos, IList parent)
        {
            if (UseSameDataView(pos, parent))
                return this.dv.Count > 0;
            this.dv = GetChildDataView(pos, parent);
            return this.dv != null && this.dv.Count > 0;
        }


        public void ResetDataView()
        {
            reset1 = true;
        }

        private bool UseSameDataView(int pos, IList parent)
        {
            if (reset1)
            {
                reset1 = false;
                return false;
            }
            return this.dv != null && parent.Equals(this.lastParent) && pos == this.lastPos;
        }

        #region ITreeGridDataSource Members


        #endregion

        #region IDisposable Members


        #endregion
    }

    #endregion

}
